# Import dependencies
import os, sys, distro
from configparser import ConfigParser

# Import internal dependencies
from ..utilities import run_command, create_logger
from ..vault_support import read_secret
from .ssl_certificate import generate_self_signed_cert



################################################################################
# VARIABLES

__prefix__ = sys.prefix



################################################################################
# LOGGING FACILITY

LOGGER_NAME = 'luksctl_api'

# Instantiate the logger
api_logger = create_logger(luks_cryptdev_file='/etc/luks/luks-cryptdev.ini',
                           logger_name=LOGGER_NAME,
                           loggers_section='logs')



################################################################################
# FUNCTIONS

def write_api_config(luks_cryptdev_file, env_path, daemons=[], node_list='',
                     exports_list='', sudo_path='/usr/bin/sudo'):
    """Writes the API configuration to the cryptdev .ini file in the luksctl_api section.

    :param luks_cryptdev_file: Path to the cryptdev .ini file, defaults to '/etc/luks/luks-cryptdev.ini'
    :type luks_cryptdev_file: str, optional
    """
    #arguments = locals()
    #arguments.pop('luks_cryptdev_file')

    config = ConfigParser()
    config.read(luks_cryptdev_file)
    # Remove luksctl_api section if written previously
    if 'luksctl_api' in config.sections():
        config.remove_section('luksctl_api')

    config.add_section('luksctl_api')
    api_config = config['luksctl_api']

    api_config['daemons'] = ','.join(daemons)
    api_config['env_path'] = env_path
    api_config['node_list'] = ','.join(node_list)
    api_config['exports_list'] = ','.join(exports_list)
    api_config['sudo_path'] = sudo_path

    with open(luks_cryptdev_file, 'w') as f:
        config.write(f)


def read_api_config(luks_cryptdev_file, api_section):
    """Reads the api configurations from the cryptdev .ini file.

    :param luks_cryptdev_file: Path to the cryptdev .ini file
    :type luks_cryptdev_file: str
    :param api_section: API section as defined in the cryptdev .ini file
    :type api_section: str
    :raises FileNotFoundError: Raises an error if the cryptdev .ini file is not found.
    :return: Returns a dictionary containing key, value pairs for each API configuration option, i.e. daemons, wn_ips, sudo_path and env_path
    :rtype: dict
    """
    
    if os.path.exists(luks_cryptdev_file):
        # Read cryptdev ini file
        config = ConfigParser()
        config.read(luks_cryptdev_file)

        # Get configuration dictionary
        api_config = dict(config[api_section].items())
        api_config['daemons'] = api_config['daemons'].split(',')
        api_config['node_list'] = api_config['node_list'].split(',')
        api_config['exports_list'] = api_config['exports_list'].split(',')

    else:
        raise FileNotFoundError('Cryptdev ini file missing.')

    return api_config


def write_systemd_unit_file(working_directory, environment_prefix, user, group, app='master_app',
                            service_file='/etc/systemd/system/luksctl-api.service',
                            gunicorn_config_file='/etc/luks/gunicorn.conf.py'):
    """General function to write the unit file used by systemd that defines the luksctl-api service.
    It's used by both master and wn classes to configure the master and wn API respectively.

    :param working_directory: Path to the luksctl_api subpackage
    :type working_directory: str
    :param environment_prefix: Path to the virtual environment in which pyluks is installed.
    :type environment_prefix: str
    :param user: User under which the luksctl-api service is run.
    :type user: str
    :param group: User group under which the luksctl-api service is run.
    :type group: str
    :param app: Flask app name as defined in the app.py module of luksctl_api, possible values are master_app or wn_app.
    :type app: str
    :param service_file: Path to the unit file, defaults to '/etc/systemd/system/luksctl-api.service'
    :type service_file: str, optional
    :param gunicorn_config_file: Path where the gunicorn.conf file is copyed, defaults to '/etc/luks/gunicorn.conf.py'
    :type gunicorn_config_file: str, optional
    """
    
    # Exit if command is not run as root
    if not os.geteuid() == 0:
        sys.exit('Error: write_systemd_unit_file must be run as root.')
    
    config = ConfigParser()
    config.optionxform = str
    
    config.add_section('Unit')
    config['Unit']['Description'] = 'Gunicorn instance to serve luksctl api server'
    config['Unit']['After'] = 'network.target'

    config.add_section('Service')
    config['Service']['User'] = user
    config['Service']['Group'] = group
    config['Service']['WorkingDirectory'] = working_directory
    config['Service']['Environment'] = f'"PATH={environment_prefix}/bin"'
    
    config['Service']['ExecStart'] = f'{environment_prefix}/bin/gunicorn --config {gunicorn_config_file} app:{app}'
    
    config.add_section('Install')
    config['Install']['WantedBy'] = 'multi-user.target'

    with open(service_file, 'w') as sf:
        config.write(sf)


def write_exports_file(exports_list, node_list):
    """Adds lines for each export directory and worker node to configure nfs exports.

    :param exports_list: List containing the directories to be exported with nfs.
    :type exports_list: list
    :param node_list: List containing IPs of the worker nodes.
    :type node_list: list
    """
    
    with open('/etc/exports','r+') as exports_file:
        for export_dir in exports_list:
            for node in node_list:
                exports_line = f'{export_dir} {node}(rw,sync,no_root_squash)'
                if exports_line not in exports_file.readlines():
                    exports_file.write(f'{export_dir} {node}(rw,sync,no_root_squash)')



################################################################################
# NODES CLASSES

class master:
    """Master node class, used to manage the luksctl API on a single virtual machine or on the master node of a cluster.
    """


    def __init__(self, luks_cryptdev_file, api_section='luksctl_api'):
        """Instantiates the master class. If luks_cryptdev_file (and eventually also api_section) is defined, other arguments will be ignored and
        the object's attributes are read from the cryptdev .ini file.

        :param luks_cryptdev_file: Path to the cryptdev .ini file, defaults to None
        :type luks_cryptdev_file: str
        :param api_section: API section as defined in the cryptdev .ini file, defaults to 'luksctl_api'
        :type api_section: str, optional
        """
        
        api_configs = read_api_config(luks_cryptdev_file=luks_cryptdev_file, api_section=api_section)
        self.daemons = api_configs['daemons']
        self.node_list = api_configs['node_list']
        self.exports_list = api_configs['exports_list']
        self.sudo_path = api_configs['sudo_path']
        self.env_path = api_configs['env_path']

        self.luksctl_cmd = f'{self.env_path}/bin/luksctl'
        self.distro_id = distro.id()


    def get_daemons(self): return self.daemons
    def get_node_list(self): return self.node_list
    def get_sudo_path(self): return self.sudo_path
    def get_env_path(self): return self.env_path


    def get_status(self):
        """Gets cryptdevice status with the 'luksctl status' command and returns a json with the following structure:
        
        * {'volume_state' : 'mounted'} if the volume is mounted.
        * {'volume_state': 'unmounted'} if the volume is unmounted.
        * {'volume_state' : 'unavailable', 'output' : stdout, 'stderr' : stderr} if the luksctl status command returns a non-recognized exit code.

        :return: String containing the json-formatted message for the volume_state
        :rtype: str
        """

        status_command = f'{self.sudo_path} {self.luksctl_cmd} status'
        stdout, stderr, status = run_command(status_command)

        api_logger.debug(f'Volume status stdout: {stdout}')
        api_logger.debug(f'Volume status stderr: {stderr}')
        api_logger.debug(f'Volume status: {status}')

        if str(status) == '0':
            return {'volume_state': 'mounted' }
        elif str(status)  == '1':
            return {'volume_state': 'unmounted' }
        else:
            return {'volume_state': 'unavailable', 'output': stdout, 'stderr': stderr }


    def open(self, vault_url, wrapping_token, secret_root, secret_path, secret_key):
        """Reads the passphrase from HashiCorp Vault, opens and mount the cryptdevice. If the master node is
        in a cluster, it restarts the nfs using the master.nfs_restart method.
        It returns a json-formatted string containing information about the cryptdevice status, refer to the
        master.get_status method for its content.

        :param vault_url: URL to Vault server
        :type vault_url: str
        :param wrapping_token: Wrapping token used to write the passphrase to Vault
        :type wrapping_token: str
        :param secret_root: Vault root in which secrets are stored, e.g. 'secrets'
        :type secret_root: str
        :param secret_path: Vault path in which the passphrase is stored.
        :type secret_path: str
        :param secret_key: Vault key associated to the passphrase.
        :type user_key: str
        :return: String containing the json-formatted message for the volume_state
        :rtype: str
        """
        
        status_command = f'{self.sudo_path} {self.luksctl_cmd} status'
        stdout, stderr, status = run_command(status_command)

        if str(status) == '0':
            return {'volume_state': 'mounted'}
        
        else:
            # Read passphrase from vault
            secret = read_secret(vault_url=vault_url,
                                 wrapping_token=wrapping_token,
                                 secret_root=secret_root,
                                 secret_path=secret_path,
                                 secret_key=secret_key)
            
            # Stop daemons before opening volume
            if self.daemons:
                self.stop_daemons()

            # Open volume
            api_logger.debug(f'Opening volume')
            open_command = f'printf "{secret}\n" | {self.sudo_path} {self.luksctl_cmd} open' 
            stdout, stderr, status = run_command(open_command)

            api_logger.debug(f'Volume status stdout: {stdout}')
            api_logger.debug(f'Volume status stderr: {stderr}')
            api_logger.debug(f'Volume status: {status}')

            if str(status) == '0':
                if self.daemons:
                    self.start_daemons()
                return {'volume_state': 'mounted' }

            elif str(status)  == '1':
                return {'volume_state': 'unmounted' }

            else:
                return {'volume_state': 'unavailable', 'output': stdout, 'stderr': stderr}


    def stop_daemons(self):
        "Stop daemons that have to be stopped before opening the volume"

        for daemon in self.daemons:
            api_logger.debug(f'Stopping {daemon}')

            stop_command = f'{self.sudo_path} systemctl stop {daemon}'
            api_logger.debug(stop_command)
            stdout, stderr, status = run_command(stop_command)

            api_logger.debug(f'{daemon} status: {status}')
            api_logger.debug(f'{daemon} status stdout: {stdout}')
            api_logger.debug(f'{daemon} status stderr: {stderr}')

    def start_daemons(self):
        "Start deamons after opening the volume"

        for daemon in self.daemons:
            api_logger.debug(f'Starting {daemon}')

            start_command = f'{self.sudo_path} systemctl start {daemon}'
            api_logger.debug(start_command)
            stdout, stderr, status = run_command(start_command)

            api_logger.debug(f'{daemon} status: {status}')
            api_logger.debug(f'{daemon} status stdout: {stdout}')
            api_logger.debug(f'{daemon} status stderr: {stderr}')




# Import dependencies
import hvac


#____________________________________
def write_secret_to_vault(vault_url, wrapping_token, secret_path, key, value, secret_root='secrets'):
    """Writes the passhprase to HashiCorp Vault.

    :param vault_url: URL to Vault server
    :type vault_url: str
    :param wrapping_token: Wrapping token used to write the passphrase to Vault.
    :type wrapping_token: str
    :param secret_path: Vault path in which the passphrase is stored.
    :type secret_path: str
    :param key: Vault key associated to the passphrase
    :type key: str
    :param value: Passphrase to be stored in Vault
    :type value: str
    """
    # Instantiate the hvac.Client class
    vault_client = hvac.Client(vault_url, verify=False)

    # Login directly with the wrapped token
    vault_client.auth_cubbyhole(wrapping_token)
    assert vault_client.is_authenticated()

    # Post secret
    secret={key:value}
    vault_client.secrets.kv.v2.create_or_update_secret(path=secret_path, secret=secret, mount_point=secret_root, cas=0)

    # Logout and revoke current token
    vault_client.logout(revoke_token=True)


#____________________________________
def read_secret(vault_url, wrapping_token, secret_root, secret_path, secret_key):
    """Read the passphrase from HashiCorp Vault.

    :param vault_url: URL to Vault server
    :type vault_url: str
    :param wrapping_token: Wrapping token used to write the passphrase to Vault.
    :type wrapping_token: str
    :param secret_root: Vault root in which secrets are stored, e.g. 'secrets'
    :type secret_root: str
    :param secret_path: Vault path in which the passphrase is stored.
    :type secret_path: str
    :param secret_key: Vault key associated to the passphrase.
    :type user_key: str
    :return: Passphrase retrieved from Vault.
    :rtype: str
    """
    
    # Instantiate the hvac.Client class
    vault_client = hvac.Client(vault_url, verify=False)

    # Login directly with the wrapped token
    vault_client.auth_cubbyhole(wrapping_token)
    assert vault_client.is_authenticated()

    # Read secret
    read_response = vault_client.secrets.kv.read_secret_version(path=secret_path, mount_point=secret_root)
    secret = read_response['data']['data'][secret_key]

    # Logout and revoke current token
    vault_client.logout(revoke_token=True)

    return secret



# Import dependencies
import random
from string import ascii_letters, digits, ascii_lowercase
import os
import sys
from pathlib import Path
from datetime import datetime
import re
import distro
from configparser import ConfigParser

# Import internal dependencies
from ..utilities import run_command, create_logger, DEFAULT_LOGFILES
from ..vault_support import write_secret_to_vault



################################################################################
# VARIABLES

alphanum = ascii_letters + digits
#now = datetime.now().strftime('-%b-%d-%y-%H%M%S')
# Get Distribution
# Ubuntu, centos, rocky currently supported
def check_distro(function):
    """Decorator function to check that the wrapped function is run on Ubuntu or CentOS.

    :param function: Function to be wrapped.
    :type function: function
    :raises Exception: Raises an exception when the wrapped function is not run on Ubuntu or CentOS
    :return: Wrapper function
    :rtype: function
    """
    global DISTNAME
    DISTNAME = distro.id()
    def wrapper_function(*args, **kwargs):
        if DISTNAME not in ['ubuntu','centos','rocky']:
            raise Exception('Distribution not supported: Ubuntu, CentOS 7, and RockyLinux 9 currently supported')
        return function()
    return wrapper_function



################################################################################
# LOGGING FACILITY

LOGGER_NAME = 'fastluks'

# Instantiate the logger
fastluks_logger = create_logger(luks_cryptdev_file='/etc/luks/luks-cryptdev.ini',
                                logger_name=LOGGER_NAME,
                                loggers_section='logs')



################################################################################
# FUNCTIONS

class LUKSError(Exception):
    pass



#____________________________________
# Volume encryption and setup functions
def create_random_cryptdev_name(n=8):
    """Generates a random string of ascii lowercase characters used as cryptdev name

    :param n: Length of the string, defaults to 8
    :type n: int
    :return: Random string of n characters
    :rtype: str
    """
    return ''.join([random.choice(ascii_lowercase) for i in range(n)])


@check_distro
def install_cryptsetup(logger=None):
    """Install the cryptsetup command line tool, used to interface with dm-crypt for creating,
    accessing and managing encrypted devices. It uses either apt or yum depending on the Linux distribution.

    :param logger: Logger object used to log information about the installation of cryptsetup, defaults to None
    :type logger: logging.Logger, optional
    """
    if DISTNAME == 'ubuntu':
        fastluks_logger.info('Distribution: Ubuntu. Using apt.')
        run_command('apt-get install -y cryptsetup pv', logger)
    else:
        fastluks_logger.info('Distribution: CentOS or RockyLinux. Using yum.')
        run_command('yum install -y cryptsetup-luks pv', logger)


@check_distro
def check_cryptsetup():
    """Checks if the dm-crypt module and cryptsetup are installed.
    """
    fastluks_logger.info('Check if the required applications are installed...')
    
    _, _, dmsetup_status = run_command('type -P dmsetup &>/dev/null')
    if dmsetup_status != 0:
        fastluks_logger.info('dmsetup is not installed. Installing...')
        if DISTNAME == 'ubuntu':
            run_command('apt-get install -y dmsetup')
        else:
            run_command('yum install -y device-mapper')
    else:
        fastluks_logger.info('dmsetup is already installed.')
    
    _, _, cryptsetup_status = run_command('type -P cryptsetup &>/dev/null')
    if cryptsetup_status != 0:
        fastluks_logger.info('cryptsetup is not installed. Installing...')
        install_cryptsetup(logger=fastluks_logger)
        fastluks_logger.info('cryptsetup installed.')
    else:
        fastluks_logger.info('cryptsetup is already installed.')


def create_random_secret(passphrase_length):
    """Creates a random passphrase of alphanumeric characters.

    :param passphrase_length: Passphrase length
    :type passphrase_length: int
    :return: Alphanumeric string of the specified length
    :rtype: str
    """
    return ''.join([random.choice(alphanum) for i in range(passphrase_length)])


def end_encrypt_procedure(SUCCESS_FILE):
    """Sends a signal to unlock waiting condition, writing 'LUKS encryption completed' in the specified
    success file. This file is used by automation software (e.g. Ansible) to make sure that the encryption
    procedure is completed.

    :param SUCCESS_FILE: Path to the encryption success file to be written.
    :type SUCCESS_FILE: str
    """
    with open(SUCCESS_FILE, 'w') as success_file:
        success_file.write('LUKS encryption completed.') # WARNING DO NOT MODFIFY THIS LINE, THIS IS A CONTROL STRING FOR ANSIBLE
    fastluks_logger.info('SUCCESSFUL.')


def end_volume_setup_procedure(SUCCESS_FILE):
    """Sends signal to unlock waiting condition, writing 'LUKS setup completed' in the specified success
    file. This file is used by automation software (e.g. Ansible) to make sure that the setup procedure
    is completed.

    :param SUCCESS_FILE: Path to the setup success file to be written.
    :type SUCCESS_FILE: str
    """
    with open(SUCCESS_FILE,'w') as success_file:
        success_file.write('Volume setup completed.') # WARNING DO NOT MODFIFY THIS LINE, THIS IS A CONTROL STRING FOR ANSIBLE
    fastluks_logger.info('SUCCESSFUL.')


def read_ini_file(cryptdev_ini_file):
    """Reads the cryptdev .ini file. Returns a dictionary containing the information of the encrypted
    device written in the .ini file.

    :param cryptdev_ini_file: Path to the cryptdev .ini file
    :type cryptdev_ini_file: str
    :return: Dictionary containing informations about the encrypted device in key-value pairs
    :rtype: dict
    """
    config = ConfigParser()
    config.read_file(open(cryptdev_ini_file))
    luks_section = config['luks']
    return {key:luks_section[key] for key in luks_section}



################################################################################
# DEVICE CLASSE

class device:
    """Device class used to create, access and manage encrypted devices. 
    """


    def __init__(self, device_name, cryptdev, mountpoint, filesystem,
                 cipher_algorithm='aes-xts-plain64', keysize=256, hash_algorithm='sha256'):
        """Instantiate a device object

        :param device_name: Name of the volume, e.g. /dev/vdb
        :type device_name: str
        :param cryptdev: Name of the cryptdevice, e.g. crypt
        :type cryptdev: str
        :param mountpoint: Mountpoint for the encrypted device, e.g. /export
        :type mountpoint: str
        :param filesystem: Filesystem for the volume, e.g. ext4
        :type filesystem: str
        :param cipher_algorithm: Algorithm for the encryption, e.g. aes-xts-plain64
        :type cipher_algorithm: str
        :param keysize: Key-size for the cipher algorithm, e.g. 256
        :type keysize: int
        :param hash_algorithm: Hash algorithm used for key derivaiton, e.g. sha256
        :type hash_algorithm: int
        """
        self.device_name = device_name
        self.cryptdev = cryptdev
        self.mountpoint = mountpoint
        self.filesystem = filesystem
        self.cipher_algorithm = cipher_algorithm
        self.keysize = keysize
        self.hash_algorithm = hash_algorithm

    def check_vol(self):
        """Checks if the mountpoint already has a volume mounted to it and if the device_name
        specified in the device object is a volume.

        :return: False if the device_name of the device object doesn't correspond to block device
        :rtype: bool
        """
        fastluks_logger.debug('Checking storage volume.')

        # Check if a volume is already mounted to mountpoint
        if os.path.ismount(self.mountpoint):
            mounted_device, _, _ = run_command(f'df -P {self.mountpoint} | tail -1 | cut -d" " -f 1')
            fastluks_logger.debug(f'Device name: {mounted_device}')

        else:
            # Check if device_name is a volume
            if Path(self.device_name).is_block_device():
                fastluks_logger.debug(f'External volume on {self.device_name}. Using it for encryption.')
                if not os.path.isdir(self.mountpoint):
                    fastluks_logger.debug(f'Creating {self.mountpoint}')
                    os.makedirs(self.mountpoint, exist_ok=True)
                    fastluks_logger.debug(f'Device name: {self.device_name}')
                    fastluks_logger.debug(f'Mountpoint: {self.mountpoint}')
            else:
                fastluks_logger.error('Device not mounted, exiting! Please check logfile:')
                fastluks_logger.error(f'No device mounted to {self.mountpoint}')
                run_command('df -h', logger=fastluks_logger)
                raise LUKSError('Volume checks not satisfied') # unlock and terminate process


    def is_encrypted(self):
        """Checks if the device is encrypted.

        :return: True if the volume is encrypted, otherwise False.
        :rtype: bool
        """
        fastluks_logger.debug('Checking if the volume is already encrypted.')
        devices, _, _ = run_command('lsblk -p -o NAME,FSTYPE')
        if re.search(f'{self.device_name}\s+crypto_LUKS', devices):
                fastluks_logger.debug('The volume is already encrypted')
                return True
        else:
            return False


    def umount_vol(self):
        """Unmount the device
        """
        fastluks_logger.debug('Umounting device.')
        run_command(f'umount {self.mountpoint}', logger=fastluks_logger)
        fastluks_logger.debug(f'{self.device_name} umounted, ready for encryption!')


    def luksFormat(self, s3cret):
        """Sets up a the device in LUKS encryption mode: sets up the LUKS device header and encrypts
        the passphrase with the indidcated cryptographic options. 

        :param s3cret: Passphrase for the encrypted volume.
        :type s3cret: str
        :param cipher_algorithm: Algorithm for the encryption, e.g. aes-xts-plain64
        :type cipher_algorithm: str
        :param keysize: Key-size for the cipher algorithm, e.g. 256
        :type keysize: int
        :param hash_algorithm: Hash algorithm used for key derivation, e.g. sha256
        :type hash_algorithm: str
        :return: A tuple containing stdout, stderr and status of the cryptsetup luksFormat command.
        :rtype: tuple
        """
        return run_command(f'printf "{s3cret}\n" | cryptsetup -v --cipher {self.cipher_algorithm} --key-size {self.keysize} --hash {self.hash_algorithm} --iter-time 2000 --use-urandom luksFormat {self.device_name} --batch-mode')


    def luksHeaderBackup(self, luks_header_backup_file):
        """Stores a binary backup of the device's LUKS header and keyslot area in the specified directory and file.


        :param luks_header_backup_file: File in which the header and keyslot area are stored.
        :type luks_header_backup_file: str
        :return: A tuple containing stdout, stderr and status of the cryptsetup luksFormat command.
        :rtype: tuple
        """
        return run_command(f'cryptsetup luksHeaderBackup --header-backup-file {luks_header_backup_file} {self.device_name}')


    def luksOpen(self, s3cret):
        """Opens the encrypted device.

        :param s3cret: Passphrase to open the encrypted device.
        :type s3cret: str
        :return: A tuple containing stdout, stderr and status of the cryptsetup luksOpen command 
        :rtype: tuple
        """
        return run_command(f'printf "{s3cret}\n" | cryptsetup luksOpen {self.device_name} {self.cryptdev}')


    def info(self):
        """Logs to stdout device informations and cryptographic options.

        :param cipher_algorithm: Algorithm for the encryption
        :type cipher_algorithm: str
        :param hash_algorithm: Hash algorithm used for key derivation
        :type hash_algorithm: str
        :param keysize: Key-size for the cipher algorithm
        :type keysize: int
        """
        fastluks_logger.info(f'LUKS header information for {self.device_name}')
        fastluks_logger.info(f'Cipher algorithm: {self.cipher_algorithm}')
        fastluks_logger.info(f'Hash algorithm {self.hash_algorithm}')
        fastluks_logger.info(f'Keysize: {self.keysize}')
        fastluks_logger.info(f'Device: {self.device_name}')
        fastluks_logger.info(f'Crypt device: {self.cryptdev}')
        fastluks_logger.info(f'Mapper: /dev/mapper/{self.cryptdev}')
        fastluks_logger.info(f'Mountpoint: {self.mountpoint}')
        fastluks_logger.info(f'File system: {self.filesystem}')

    def setup_device(self, luks_header_backup_file, passphrase_length, passphrase, use_vault, vault_url, wrapping_token, secret_path, user_key):
        """Performs the setup wrokflow to encrypt the device by performing the following steps:

        * Logs to stdout device informations and cryptographic options with the device.info method
        * Checks the specified passphrase or creates a new one of the specified length
        * Sets up a the device in LUKS encryption mode with the device.luksFormat method
        * Stores the passphrase to HashiCorp Vault if `use_vault` is set to True with the write_secret_to_vault function.
        * Stores the header backup with the device.luksHeaderBackup method

        It either returns the passphrase if the setup is successful or False if it fails.

        :param luks_header_backup_file: File in which the header and keyslot area are stored.
        :type luks_header_backup_file: str
        :param passphrase_length: Lenght of the passphrase to be generated.
        :type passphrase_length: int
        :param passphrase: Specified passphrase to be used for device encryption.
        :type passphrase: str
        :param use_vault: If set to True, the passphrase is stored to HashiCorp Vault.
        :type use_vault: bool
        :param vault_url: URL of Vault server. 
        :type vault_url: str
        :param wrapping_token: Wrapping token used to write the passphrase on Vault.
        :type wrapping_token: str
        :param secret_path: Vault path in which the passphrase is stored.
        :type secret_path: str
        :param user_key: Vault key associated to the passphrase.
        :type user_key: str
        :return: The passphrase if the setup is successful or False if it fails.
        :rtype: str or bool
        """
        fastluks_logger.info('Start the encryption procedure.')
        fastluks_logger.debug(f'Using {self.cipher_algorithm} algorithm to luksformat the volume.')
        fastluks_logger.debug('Start cryptsetup')
        self.info()
        fastluks_logger.debug('Cryptsetup full command:')
        fastluks_logger.debug(f'cryptsetup -v --cipher {self.cipher_algorithm} --key-size {self.keysize} --hash {self.hash_algorithm} --iter-time 2000 --use-urandom --verify-passphrase luksFormat {device} --batch-mode')

        if passphrase_length == None:
            if passphrase == None:
                fastluks_logger.error("Missing passphrase!")
                raise LUKSError('Device setup procedure failed.') # unlock and exit
            s3cret = passphrase
        else:
            s3cret = create_random_secret(passphrase_length)
        
        # Start encryption procedure
        self.luksFormat(s3cret)

        # Write the secret to vault
        if use_vault:
            write_secret_to_vault(vault_url, wrapping_token, secret_path, user_key, s3cret)
            fastluks_logger.info('Passphrase stored in Vault')

        # Backup LUKS header
        luks_header_backup_dir = os.path.dirname(luks_header_backup_file)
        if not os.path.isdir(luks_header_backup_dir):
            os.mkdir(luks_header_backup_dir)
        _, _, luksHeaderBackup_ec = self.luksHeaderBackup(luks_header_backup_file)

        if luksHeaderBackup_ec != 0:
            # Cryptsetup returns 0 on success and a non-zero value on error.
            # Error codes are:
            # 1 wrong parameters
            # 2 no permission (bad passphrase)
            # 3 out of memory
            # 4 wrong device specified
            # 5 device already exists or device is busy.
            fastluks_logger.error(f'Command cryptsetup failed with exit code {luksHeaderBackup_ec}! Mounting {self.device_name} to {self.mountpoint} and exiting.')
            if luksHeaderBackup_ec == 2:
                fastluks_logger.error('Bad passphrase. Please try again.')
            raise LUKSError('Device setup procedure failed.') # unlock and exit

        return s3cret

    def open_device(self, s3cret):
        """Opens and mounts the encrypted device.

        :param s3cret: Passphrase to open the encrypted device.
        :type s3cret: str
        :return: False if any error occur (e.g. if the passphrase is wrong or if the crypt device already exists) 
        :rtype: bool, optional
        """
        if not Path(f'/dev/mapper/{self.cryptdev}').is_block_device():
            fastluks_logger.info(f'Opening LUKS volume and mapping it to /dev/mapper/{self.cryptdev}')
            _, _, openec = self.luksOpen(s3cret)
            
            if openec != 0:
                if openec == 2:
                    fastluks_logger.error('Bad passphrase. Please try again.')
                    raise LUKSError('luksOpen failed, mapping not created.') # unlock and exit
                else:
                    fastluks_logger.error(f'Crypt device already exists! Please check logs: {LOGFILE}')
                    fastluks_logger.error('Unable to luksOpen device.')
                    fastluks_logger.error(f'/dev/mapper/{self.cryptdev} already exists.')
                    fastluks_logger.error(f'Mounting {self.device_name} to {self.mountpoint} again.')
                    run_command(f'mount {self.device_name} {self.mountpoint}', logger=fastluks_logger)
                    raise LUKSError('luksOpen failed, mapping not created.') # unlock and exit
        else:
            fastluks_logger.info(f'LUKS volume already opened and mapped to /dev/mapper/{self.cryptdev}')


    def encryption_status(self):
        """Checks cryptdevice status, with the command cryptsetup status. It logs stdout, stderr
        and status to the logfile.
        """
        fastluks_logger.debug(f'Check {self.cryptdev} status with cryptsetup status')
        run_command(f'cryptsetup -v status {self.cryptdev}', logger=fastluks_logger)


    def create_cryptdev_ini_file(self, luks_cryptdev_file, luks_header_backup_file,
                                 save_passphrase_locally, s3cret):
        """Creates the cryptdev .ini file containing information of the encrypted device under the 'luks' section.
        It also stores the default paths for the log files of fastluks, luksctl and luksctl_api subpackages in the 'logs' section.
        After creating the ini file, it logs the output of 'dmsetup info' and 'cryptsetup luksDump' commands.

        :param luks_cryptdev_file: Path to the cryptdev .ini file.
        :type luks_cryptdev_file: str
        :param cipher_algorithm: Algorithm for the encryption, e.g. aes-xts-plain64
        :type cipher_algorithm: str
        :param hash_algorithm: Hash algorithm used for the key derivation, e.g. sha256
        :type hash_algorithm: str
        :param keysize: Key-size for the cipher algorithm, e.g. 256
        :type keysize: int
        :param luks_header_backup_file: File in which the header and keyslot area are stored.
        :type luks_header_backup_file: str
        :param save_passphrase_locally: If set to true, the passphrase is written in the .ini file in plain text. This option is usually used for testing purposes.
        :type save_passphrase_locally: bool
        :param s3cret: Passphrase to open the encrypted device, written in the .ini file only if `save_passphrase_locally` is set to True
        :type s3cret: str
        """
        luksUUID, _, _ = run_command(f'cryptsetup luksUUID {self.device_name}')
        luksUUID = luksUUID.rstrip()

        with open(luks_cryptdev_file, 'w') as f:
            config = ConfigParser()
            config.add_section('luks')
            config_luks = config['luks']
            config_luks['cipher_algorithm'] = self.cipher_algorithm
            config_luks['hash_algorithm'] = self.hash_algorithm
            config_luks['keysize'] = str(self.keysize)
            config_luks['device'] = self.device_name
            config_luks['uuid'] = luksUUID
            config_luks['cryptdev'] = self.cryptdev
            config_luks['mapper'] = f'/dev/mapper/{self.cryptdev}'
            config_luks['mountpoint'] = self.mountpoint
            config_luks['filesystem'] = self.filesystem
            config_luks['header_path'] = f'{luks_header_backup_file}'

            config.add_section('logs')
            config_logs = config['logs']
            for name,logfile in DEFAULT_LOGFILES.items():
                config_logs[name] = logfile

            if save_passphrase_locally:
                config_luks['passphrase'] = s3cret
                config.write(f)
                fastluks_logger.info(f'Device informations and key have been saved in {luks_cryptdev_file}')
            else:
                config.write(f)
                fastluks_logger.info(f'Device informations have been saved in {luks_cryptdev_file}')

        run_command(f'dmsetup info /dev/mapper/{self.cryptdev}', logger=fastluks_logger)
        run_command(f'cryptsetup luksDump {self.device_name}', logger=fastluks_logger)


    def wipe_data(self):
        """Paranoid mode function: it wipes the disk by overwriting the entire drive with random data.
        It may take some time.
        """
        fastluks_logger.info('Paranoid mode selected. Wiping disk')
        fastluks_logger.info('Wiping disk data by overwriting the entire drive with random data.')
        fastluks_logger.info('This might take time depending on the size & your machine!')
        
        run_command(f'dd if=/dev/zero of=/dev/mapper/{self.cryptdev} bs=1M status=progress')
        
        fastluks_logger.info(f'Block file /dev/mapper/{self.cryptdev} created.')
        fastluks_logger.info('Wiping done.')

    def create_fs(self):
        """Creates the filesystem for the LUKS encrypted device based on the `filesystem` attribute of the device object.

        :return: False if the mkfs command fails.
        :rtype: False, optional
        """
        fastluks_logger.info('Creating filesystem.')
        fastluks_logger.debug(f'Creating {self.filesystem} filesystem on /dev/mapper/{self.cryptdev}')
        _, _, mkfs_ec = run_command(f'mkfs -t {self.filesystem} /dev/mapper/{self.cryptdev}', logger=fastluks_logger)
        if mkfs_ec != 0:
            fastluks_logger.error(f'While creating {self.filesystem} filesystem. Please check logs.')
            fastluks_logger.error('Command mkfs failed!')
            raise LUKSError('Command mkfs failed.') # unlock and exit


    def mount_vol(self):
        """Mounts the encrypted device to the 'mountpoint' specified in the device attributes.
        """
        fastluks_logger.info('Mounting encrypted device.')
        fastluks_logger.debug(f'Mounting /dev/mapper/{self.cryptdev} to {self.mountpoint}')
        run_command(f'mount /dev/mapper/{self.cryptdev} {self.mountpoint}', logger=fastluks_logger)
        run_command('df -Hv', logger=fastluks_logger)


    def encrypt(self, luks_header_backup_file, luks_cryptdev_file,
                passphrase_length, passphrase, save_passphrase_locally,
                use_vault, vault_url, wrapping_token, secret_path, user_key):
        """Performs the encryption workflow with the following steps:

        * Creates the lock file with the lock function.
        * Creates a random name for the cryptdevice with the create_random_cryptdev_name function.
        * Checks the device volume with the device.check_vol method.
        * Checks if the volume is encrypted with the device.is_encrypted method, if it's not starts the encryption.
        * Encryption procedure:
            * Unmount the volume with the device.umount_vol method
            * Performs the device setup workflow for encryption with the device.setup_device method.
            * Unlocks and exits if the setup procedure failed.
        * Opens the successfully encrypted device with the device.open_device method, unlocks and exits if it fails.
        * Checks the encryption status with the device.encryption_status method.
        * Creates the cryptdev .ini file with the device information with the device.create_cryptdev_ini_file method.
        * Ends the encryption procedure with the end_encrypt_procedure function.
        * Unlocks and exits.

        :param cipher_algorithm: Algorithm for the encryption, e.g. aes-xts-plain64
        :type cipher_algorithm: str
        :param keysize: Key-size for the cipher algorithm, e.g. 256
        :type keysize: int
        :param hash_algorithm: Hash algorithm used for the key derivation, e.g. sha256
        :type hash_algorithm: str
        :param luks_header_backup_file: File in which the header and keyslot area are stored.
        :type luks_header_backup_file: str
        :param LOCKFILE: Path to the lockfile.
        :type LOCKFILE: str
        :param SUCCESS_FILE: Path to the encryption success file.
        :type SUCCESS_FILE: str
        :param luks_cryptdev_file: Path to the cryptdev .ini file.
        :type luks_cryptdev_file: str
        :param passphrase_length: Length of the passphrase to be generated.
        :type passphrase_length: int
        :param passphrase: Specified passphrase to be used for device encryption.
        :type passphrase: str
        :param save_passphrase_locally: If set to true, the passphrase is written in the cryptdev .ini file.
        :type save_passphrase_locally: bool
        :param use_vault: If set to true, the passphrase is stored to HashiCorp Vault.
        :type use_vault: bool
        :param vault_url: URL of Vault server.
        :type vault_url: str
        :param wrapping_token: Wrapping token used to write the passphrase on Vault.
        :type wrapping_token: str
        :param secret_path: Vault path in which the passhprase is stored.
        :type secret_path: str
        :param user_key: Vault key associated to the passphrase.
        :type user_key: str
        """
        
        cryptdev = create_random_cryptdev_name() # Assign random name to cryptdev

        check_cryptsetup() # Check that cryptsetup and dmsetup are installed

        self.check_vol() # Check which virtual volume is mounted to mountpoint, unlock and exit if it's not mounted

        if not self.is_encrypted(): # Check if the volume is encrypted, if it's not start the encryption procedure
            self.umount_vol()
            s3cret = self.setup_device(luks_header_backup_file, passphrase_length, passphrase,
                                       use_vault, vault_url, wrapping_token, secret_path, user_key)
        else:
            raise LUKSError('Device is already encrypted')

        self.open_device(s3cret) # Create mapping

        self.encryption_status() # Check status

        self.create_cryptdev_ini_file(luks_cryptdev_file, luks_header_backup_file, save_passphrase_locally, s3cret) # Create ini file


    def volume_setup(self):
        """Performs the setup workflow for the encrypted volume with the following steps:

        * Creates a lockfile with the lock function.
        * Creates the encrypted volume filesystem with the device.create_fs method. Unlocks and exits if it fails.
        * Mounts the encrypted volume.
        * Creates the setup success file with the end_volume_setup_procedure function.
        * Unlocks

        :param LOCKFILE: Path to the lockfile.
        :type LOCKFILE: str
        :param SUCCESS_FILE: Path to the volume setup success file.
        :type SUCCESS_FILE: str
        """
        
        self.create_fs() # Create filesystem

        self.mount_vol() # Mount volume



# Import dependencies
import subprocess
import os
from configparser import ConfigParser
import logging
import sys



################################################################################
# VARIABLES

DEFAULT_LOGFILES = {
    'fastluks':'/tmp/fastluks.log',
    'luksctl':'/tmp/luksctl.log',
    'luksctl_api':'/tmp/luksctl-api.log'
}



################################################################################
# FUNCTIONS

#__________________________________
# Function to run bash commands
def run_command(cmd, logger=None):
    """Run subprocess call redirecting stdout, stderr and the command exit code.

    :param cmd: Command to be executed.
    :type cmd: str
    :param logger: logging.Logger object used to log stdout, stderr and exit code, defaults to None
    :type logger: loggin.Logger, optional
    :return: Returns tuple containing stdout, stderr and exit code.
    :rtype: tuple
    """
    proc = subprocess.Popen(args=cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    communicateRes = proc.communicate()
    stdout, stderr = [x.decode('utf-8') for x in communicateRes]
    status = proc.wait()

    # Functionality to replicate cmd >> "$LOGFILE" 2>&1
    if logger != None:
        logger.debug(f'Command: {cmd}\nStdout: {stdout}\nStderr: {stderr}')
    
    return stdout, stderr, status


#__________________________________
# Create logging facility
def create_logger(luks_cryptdev_file, logger_name, loggers_section='logs'):
    """Instantiate a logging.Logger object which logs to the file specified in the cryptdev .ini file.

    :param luks_cryptdev_file: Path to the cryptdev .ini file containing the path to the log file.
    :type luks_cryptdev_file: str
    :param logger_name: Logger name assigned to the logging.Logger object.
    :type logger_name: str
    :param loggers_section: Loggers section as defined in the cryptdev .ini file, defaults to 'logs'
    :type loggers_section: str, optional
    :return: logging.Logger object.
    :rtype: loggin.Logger
    """
    # Read the logfile path from the ini file (or use the default logfile if ini file missing)
    logfile = get_logfile(luks_cryptdev_file=luks_cryptdev_file,
                          logger_name=logger_name,
                          loggers_section=loggers_section)
    
    # Create logfile if it doesn't exist
    if not os.path.exists(logfile):
        create_logfile(path=logfile)

    # Define logging format
    formatter = logging.Formatter('%(levelname)s %(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    
    # Define file logging handler
    file_handler = logging.FileHandler(logfile, mode='a+')  
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)

    # Define stdout logging handler
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    stdout_handler.setLevel(logging.INFO)

    # Create logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(stdout_handler)

    return logger


def get_logfile(luks_cryptdev_file, logger_name, loggers_section='logs'):
    """Returns the path to the log file as defined in the cryptdev .ini file. If the log file is not defined
    in the .ini file, a default value is returned for the logger_name (possible values for logger_name
    are fastluks, luksctl and luksct_api)

    :param luks_cryptdev_file: Path to the cryptdev .ini file
    :type luks_cryptdev_file: str
    :param logger_name: Name for the logger, possible values are fastluks, luksctl and luksctl_api
    :type logger_name: str
    :param loggers_section: Loggers section as defined in the cryptdev .ini file, defaults to 'logs'
    :type loggers_section: str, optional
    :return: Path to the log file (either the default or the one specified in the cryptdev .ini file)
    :rtype: str
    """
    # Read logger file from cryptdev file
    if os.path.exists(luks_cryptdev_file):
        config = ConfigParser()
        config.read(luks_cryptdev_file)
        if loggers_section in config.sections():
            if logger_name in config[loggers_section]:
                logfile = config[loggers_section][logger_name]
                return logfile
    
    # cryptdev file or logger section/value missing, return default logger
    return DEFAULT_LOGFILES[logger_name]


def create_logfile(path):
    """Creates the log file with 666 permissions in the specified path.

    :param path: Path in which the log file is created
    :type path: str
    """
    with open(path, 'w+'):
            pass
    os.chmod(path, 0o666)



# import dependencies
import os, sys
from configparser import ConfigParser

# Import internal dependencies
from ..utilities import run_command, create_logger



################################################################################
# LOGGING FACILITY

LOGGER_NAME = 'luksctl'

# Instantiate the logger
luksctl_logger = create_logger(luks_cryptdev_file='/etc/luks/luks-cryptdev.ini',
                               logger_name=LOGGER_NAME,
                               loggers_section='logs')



################################################################################
# LUKSCtl class

class LUKSCtl:
    """Class that provides functions to manage a LUKS encrypted device. It reads the encrypted device information
    from the cryptdev .ini file.
    """


    def __init__(self, config_file):
        """Instantiate a LUKSCtl object.

        :param config_file: Path to the cryptdev .ini file.
        :type config_file: str
        """

        self.config_file = config_file

        config = ConfigParser()
        config.read(config_file)
        luks_config = config['luks']

        self.cipher_algorithm = luks_config['cipher_algorithm']
        self.hash_algorithm = luks_config['hash_algorithm']
        self.keysize = luks_config['keysize']
        self.device = luks_config['device']
        self.uuid = luks_config['uuid']
        self.cryptdev = luks_config['cryptdev']
        self.mapper = luks_config['mapper']
        self.mountpoint = luks_config['mountpoint']
        self.filesystem = luks_config['filesystem']


    def get_cipher_algorithm(self): return self.cipher_algorithm
    def get_hash_algorithm(self): return self.hash_algorithm
    def get_keysize(self): return self.keysize
    def get_device(self): return self.device
    def get_uuid(self): return self.uuid
    def get_cryptdev(self): return self.cryptdev
    def get_mapper(self): return self.mapper
    def get_mountpoint(self): return self.mountpoint
    def get_filesystem(self): return self.filesystem


    def set_cipher_algorithm(self, cipher_algorithm): self.cipher_algorithm = cipher_algorithm
    def set_hash_algorithm(self, hash_algorithm): self.hash_algorithm = hash_algorithm
    def set_keysize(self, keysize): keysize = keysize
    def set_device(self, device): self.device = device
    def set_uuid(self, uuid): self.uuid = uuid
    def set_cryptdev(self, cryptdev): self.cryptdev = cryptdev
    def set_mapper(self, mapper): self.mapper = mapper
    def set_mountpoint(self, mountpoint): self.mountpoint = mountpoint
    def set_filesystem(self, filesystem): self.filesystem = filesystem


    def dmsetup_info(self):
        """Runs the command 'dmsetup info' on the cryptdevice.

        :return: Returns the command exit code.
        :rtype: int
        """

        _, _, status = run_command(f'dmsetup info /dev/mapper/{self.cryptdev}')
        return status
  

    def display_dmsetup_info(self):
        """Displays the cryptdevice status. It prints the device information, followed by 'Encrypted volume: [ OK ]'
        if the device is correctly setup and open, otherwise it prints 'Encrypted volume: [ FAIL ]'.
        """

        stdOutValue, stdErrValue, status = run_command(f'dmsetup info /dev/mapper/{self.cryptdev}')

        if str(status) == '0':
            print(stdOutValue)
            print('Encrypted volume: [ OK ]')
            sys.exit(0)
        else:
            luksctl_logger.debug(f'[luksctl] {stdErrValue}')
            print('Encrypted volume: [ FAIL ]')
            sys.exit(1)
    

    def luksopen_device(self):
        """Opens the cryptdevice, mounts it and call the LUKSCtl.display_dmsetup_info method. It prints
        'Encrypted volume mount: [ FAIL ]' if the mount command returns an error.
        """

        run_command(f'cryptsetup luksOpen /dev/disk/by-uuid/{self.uuid} {self.cryptdev}')
    
        _, _, status = run_command(f'mount /dev/mapper/{self.cryptdev} {self.mountpoint}')
    
        if str(status) == '0':
            self.display_dmsetup_info()
        else:
            print('Encrypted volume mount: [ FAIL ]')
            sys.exit(1)
    

    def luksclose_device(self):
        """Unmounts, closes the cryptdevice and calls the LUKSCtl.dmsetup_info method. It prints
        'Encrypted volume umount: [ OK ]' if the cryptdevice is correctly closed, 'Encrypted volume umount [ FAIL ]'
        otherwise.
        """

        run_command(f'umount {self.mountpoint}') # Unmount device

        run_command(f'cryptsetup close {self.cryptdev}') # Close device

        # if dmsetup_setup fails (status 1) the volume has been correctly closed
        if str(self.dmsetup_info()) == '0':
            print('Encrypted volume umount: [ FAIL ]')
            sys.exit(1)
        else:
            print('Encrypted volume umount: [ OK ]')
            sys.exit(0)



# Import dependencies
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography import x509
from cryptography.hazmat.primitives import hashes
import datetime

def generate_private_key(key_size, key_file):
    """Generates a private key for the self signed certificate.

    :param key_size: Key length in bits
    :type key_size: int
    :param key_file: Path where the private key is stored.
    :type key_file: str
    :return: An instance of RSAPrivateKey containing the private key.
    :rtype: RSAPrivateKey
    """
    
    key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)

    with open(key_file, 'wb') as f:
        f.write(key.private_bytes(encoding=serialization.Encoding.PEM, format=serialization.PrivateFormat.TraditionalOpenSSL, encryption_algorithm=serialization.NoEncryption()))

    return key


def generate_self_signed_cert(CN='localhost', cert_file='/etc/luks/gunicorn-cert.pem', expiration_days=3650, key_size=4096, key_file='/etc/luks/gunicorn-key.pem'):
    """Generates a self signed certificate used by the luksctl API for https.

    :param CN: DNS name, defaults to 'localhost'
    :type CN: str, optional
    :param cert_file: Path where the certificate is stored, defaults to '/etc/luks/gunicorn-cert.pem'
    :type cert_file: str, optional
    :param expiration_days: Expiration of the self signed certificate, defaults to 3650
    :type expiration_days: int, optional
    :param key_size: Private key length in bits, defaults to 4096
    :type key_size: int, optional
    :param key_file: Path where the private key is stored, defaults to '/etc/luks/gunicorn-key.pem'
    :type key_file: str, optional
    """
    subject = issuer = x509.Name([x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, CN)])

    key = generate_private_key(key_size=key_size, key_file=key_file)

    cert = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(
    # The certificate will be valid for 3650 days
        datetime.datetime.utcnow() + datetime.timedelta(days=expiration_days)
    ).add_extension(
        x509.SubjectAlternativeName([x509.DNSName(CN)]), critical=False,
    # Sign the certificate with private key
    ).sign(key, hashes.SHA256())

    with open(cert_file, 'wb') as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))



# Import dependencies
from flask import Flask, request, abort, jsonify
import json
import os
import logging
from configparser import ConfigParser

# Import internal dependencies
from .luksctl_run import master, api_logger



################################################################################
# APP CONFIGS

app = Flask(__name__)

def instantiate_master_node():
    """Instantiate the master_node object needed by the API functions.

    :return: A master object which attributes are retrieved from the cryptdev .ini file.
    :rtype: pyluks.luksctl_api.luksctl_run.master
    """
    master_node = master(luks_cryptdev_file='/etc/luks/luks-cryptdev.ini', api_section='luksctl_api')
    return master_node



################################################################################
# FUNCTIONS

@app.route('/luksctl_api/v1.0/status', methods=['GET'])
def get_status():
    """Runs the master.get_status method on a GET request.

    :return: Output from the master.get_status method.
    :rtype: str
    """

    master_node = instantiate_master_node()
    
    response = master_node.get_status()

    return jsonify(response)


@app.route('/luksctl_api/v1.0/open', methods=['POST'])
def luksopen():
    """Runs the master.open method on a POST request containing the HashiCorp Vault informations to retrieve
    the passphrase.

    :return: Output from the master.open method.
    :rtype: str
    """

    master_node = instantiate_master_node()

    if not request.json or \
       not 'vault_url' in request.json or \
       not 'vault_token' in request.json or \
       not 'secret_root' in request.json or \
       not 'secret_path' in request.json or \
       not 'secret_key' in request.json:
       abort(400)

    wn_list = master_node.get_node_list() 
    if wn_list != None:
        api_logger.debug(wn_list)

    response = master_node.open(vault_url=request.json['vault_url'],
                                wrapping_token=request.json['vault_token'],
                                secret_root=request.json['secret_root'],
                                secret_path=request.json['secret_path'],
                                secret_key=request.json['secret_key'])
    
    return jsonify(response)



# Import dependencies
import zc.lockfile
import sys
import os
import re

LOCKFILE = '/var/run/fast-luks.lock'

#____________________________________
# Lock/UnLock Section
def lock():
    """Generate lockfile in order to avoid multiple instances to encrypt at the same time.

    :return: lockfile instance.
    :rtype: zc.lockfile.LockFile
    """
    # Start locking attempt
    try:
        LOCK = zc.lockfile.LockFile(LOCKFILE, content_template='{pid};{hostname}') # storing the PID and hostname in LOCKFILE
        return LOCK
    except zc.lockfile.LockError:
        # Lock failed: retrieve the PID of the locking process
        with open(LOCKFILE, 'r') as lock_file:
            pid_hostname = lock_file.readline()
            PID = re.search(r'^\s(\d+);', pid_hostname).group()
        echo('ERROR', f'Another script instance is active: PID {PID}')
        sys.exit(2)

    # lock is valid and OTHERPID is active - exit, we're locked!
    echo('ERROR', f'Lock failed, PID {PID} is active')
    echo('ERROR', f'Another fastluks process is active')
    echo('ERROR', f'If you are sure fastluks is not already running,')
    echo('ERROR', f'You can remove {LOCKFILE} and restart fastluks')
    sys.exit(2)


def unlock(LOCK, do_exit=True, message=None):
    """Performs the unlocking of a lockfile and terminates the process if specified.

    :param lock: LockFile object instantiated by the lock function.
    :type lock: zc.lockfile.LockFile
    :param LOCKFILE: Path to the lockfile to be unlocked.
    :type LOCKFILE: str
    :param do_exit: If set to True, the process will be terminated after the unlocking, defaults to True
    :type do_exit: bool, optional
    :param message: Message printed when the process is terminated, defaults to None
    :type message: str, optional
    """
    LOCK.close()
    os.remove(LOCKFILE)
    if do_exit:
        sys.exit(f'UNLOCK: {message}')


# def lock_process(function):
#     """Wrapper function to use a lockfile while a function is running

#     :param function: Function to wrap
#     :type function: function
#     """
#     def wrapper_function(*args, **kwargs):
#         LOCK = lock(LOCKFILE) # lock
#         result = function(*args, **kwargs)
#         unlock(LOCK, LOCKFILE, do_exit=False) # Unlock
#         return result
#     return wrapper_function


import os

workers = 2
bind = '0.0.0.0:5000'
umask = 0o007
_certfile_path = '/etc/luks/gunicorn-cert.pem'
certfile = _certfile_path if os.path.exists(_certfile_path) else None
_keyfile_path = '/etc/luks/gunicorn-key.pem'
keyfile = _keyfile_path if os.path.exists(_keyfile_path) else None
